import shutil
import tempfile
import zipfile
from pathlib import Path, PosixPath
from dataclasses import dataclass
from lxml import etree as et


@dataclass
class ScalarVariable:
    name: str
    description: str
    causality: str
    variability: str


class VariableList(list):

    def __repr__(self):
        print_out = """"""
        for item in self:
            print_out += f"""
        ___________________________________________________________________________________________
               Name: {item.name}
        Description: {item.description}
        Variability: {item.variability}
          Causality: {item.causality}
            """
        return print_out


class FMU:

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        shutil.rmtree(self.temp_dir)

    def __repr__(self):
        return f"""
        Functional Mockup Unit:
             Model Name: {self.model_name}
            FMI Version: {self.fmi_version}
               Filepath: {self.file_path}
        """

    def __init__(self, file_path):
        """This class allows for peeking into a FMU file, it by no means covers the entirety of the FMI standard and should
             primarily be considered as a debugging tool for looking of model input, outputs and parameters.
            :param file_path: filepath to the target FMU.
            :type file_path: str or PosixPath
        """
        if type(file_path) is not PosixPath:
            file_path = Path(file_path)

        self.temp_dir = tempfile.mkdtemp()
        self.file_path = file_path
        self.model_description_file = None
        self.guid = None

        self.__variables: VariableList[ScalarVariable] = VariableList()
        self.model_name = None
        self.fmi_version = None
        self.__read__()

    def __read__(self):
        with zipfile.ZipFile(self.file_path, 'r') as zip_ref:
            zip_ref.extractall(self.temp_dir)

        self.model_description_file = list(Path(self.temp_dir).glob('modelDescription.xml'))[0]
        tree = et.parse(self.model_description_file)
        root = tree.getroot()

        self.guid = root.get('guid')
        self.model_name = root.get('modelName')
        self.fmi_version = root.get('fmiVersion')

        model_variables = root.findall('ModelVariables')[0]
        scalar_variables = model_variables.findall('ScalarVariable')
        for scalar in scalar_variables:
            name = scalar.get('name')
            description = scalar.get('description')
            causality = scalar.get('causality')
            variability = scalar.get('variability')
            scalar_variable = ScalarVariable(name=name, description=description,
                                             variability=variability, causality=causality)
            self.__variables.append(scalar_variable)

    @property
    def parameters(self) -> VariableList:
        return VariableList([item for item in self.__variables if item.causality == 'parameter'])

    @property
    def outputs(self) -> VariableList:
        return VariableList([item for item in self.__variables if item.causality == 'output'])

    @property
    def inputs(self) -> VariableList:
        return VariableList([item for item in self.__variables if item.causality == 'input'])

    def exist(self, name: str):
        """ Returns true if a scalar variable exist with the given name """
        for entry in self.__variables:
            if entry.name == name:
                return True
        return False

    def get(self, causality: str = None, variability: str = None) -> VariableList:
        """
        Get a variable from the FMU, based on the attributes' causality, variability.
        :param causality: parameter, input etc.
        :param variability: fixed, tunable etc.
        :return: list of matching variables
        """
        return VariableList([item for item in self.__variables
                             if (causality is None or item.causality == causality) and
                             (variability is None or item.variability == variability)])

    def variables(self) -> VariableList:
        return self.__variables



import pathlib

import hashlib
from pyssp_standard.utils import XMLFile
from pyssp_standard.standard import SRMDStandard
from lxml import etree as et
from lxml.etree import QName


def register_namespaces():
    srmd_standards = SRMDStandard
    for name, url in srmd_standards.namespaces.items():
        et.register_namespace(name, url)


class ClassificationEntry(SRMDStandard):
    def __init__(self, keyword: str = None, content: str = None, *, element=None):
        self.keyword = keyword
        self.content = content
        if element is not None:
            self.__read__(element)

    def __read__(self, element):
        pass

    def as_element(self):
        entry = et.Element(QName(self.namespaces['stc'], 'ClassificationEntry'), attrib={'keyword': self.keyword})
        entry.text = self.content
        return entry


class Classification(SRMDStandard):

    def __init__(self, classification_type: str = None, *, element=None):
        self.__classification_entries = []
        self.__classification_type = classification_type
        if element is not None:
            self.__read__(element)

    def add_classification_entry(self, classification_entry: ClassificationEntry):
        self.__classification_entries.append(classification_entry)

    def __read__(self, element):
        pass

    def as_element(self):
        classification = et.Element(QName(self.namespaces['stc'], 'Classification'),
                                    attrib={'type': self.__classification_type})
        for entry in self.__classification_entries:
            classification.append(entry.as_element())

        return classification


class SRMD(XMLFile, SRMDStandard):

    def __init__(self, file_path, mode='r', name='unnamed'):
        self.__name = name
        self.__classifications = []
        self.data = None
        self.checksum = None
        self.checksum_type = "SHA3-256"
        self.version = "1.0.0-beta2"

        super().__init__(file_path, mode)

    def __check_compliance__(self):
        super().check_compliance(self.schema, self.namespaces)

    def assign_data(self, filepath, create_checksum=True):
        if type(filepath) is not pathlib.PosixPath:
            filepath = pathlib.Path(filepath)
        self.data = str(filepath)

        if create_checksum:
            with open(filepath) as file:
                data = file.read()
                self.checksum = hashlib.sha3_256(data.encode()).hexdigest()

    def __read__(self):
        self.__tree = et.parse(self.file_path)
        self.root = self.__tree.getroot()
        self.version = self.root.get('version')
        self.__name = self.root.get('name')
        self.data = self.root.get('data')
        self.__checksum = self.root.get('checksum')
        self.checksum_type = self.root.get('checksumType')

        self.top_level_metadata.update(self.root.attrib)
        self.base_element.update(self.root.attrib)

        classifications = self.root.findall('stc:Classification', self.namespaces)
        for classification in classifications:
            self.add_classification(Classification(element=classification))

    def __enter__(self):
        register_namespaces()
        return self

    def __write__(self):
        attributes = {'version': self.version, 'name': self.__name}
        if self.data is not None:
            attributes['data'] = self.data
        if self.checksum is not None:
            attributes['checksum'] = self.checksum
            attributes['checksumType'] = self.checksum_type
        self.root = et.Element(QName(self.namespaces['srmd'], 'SimulationResourceMetaData'), attrib=attributes)
        self.root = self.top_level_metadata.update_root(self.root)
        self.root = self.base_element.update_root(self.root)
        for classification in self.__classifications:
            self.root.append(classification.as_element())

    def add_classification(self, classification: Classification):
        self.__classifications.append(classification)




import tempfile
from pathlib import Path, PosixPath
from abc import ABC, abstractmethod
import xmlschema
from lxml import etree as ET

from pyssp_standard.common_content_ssc import Annotations, Annotation, BaseElement, TopLevelMetaData
from pyssp_standard.standard import SSPStandard


def register_namespaces():
    ssp_standards = SSPStandard
    for name, url in ssp_standards.namespaces.items():
        ET.register_namespace(name, url)


class XMLFile(ABC):

    @abstractmethod
    def __read__(self):
        pass

    @abstractmethod
    def __write__(self):
        pass

    @property
    def BaseElement(self):
        return self.base_element

    @property
    def TopLevelMetaData(self):
        return self.top_level_metadata

    def __init__(self, file_path, mode='r'):
        self.__mode = mode
        if type(file_path) is not PosixPath:
            file_path = Path(file_path)
        self.__file_path = file_path
        self.__tree = None
        self.root = None
        self.base_element: BaseElement = BaseElement()
        self.top_level_metadata: TopLevelMetaData = TopLevelMetaData()

        if mode == 'r' or mode == 'a':
            self.__read__()

    def check_compliance(self, schema, namespaces):
        if self.__mode in ['a', 'w']:  # Temporary file creation
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_file_path = temp_dir + 'tmp.xml'
                self.__write__()
                xml_string = ET.tostring(self.root, pretty_print=True, encoding='utf-8', xml_declaration=True)
                with open(temp_file_path, 'wb') as file:
                    file.write(xml_string)
                xmlschema.validate(temp_file_path, schema, namespaces=namespaces)
        else:
            xmlschema.validate(self.file_path, schema, namespaces=namespaces)

    @property
    def file_path(self):
        return self.__file_path

    def __save__(self):
        xml_string = ET.tostring(self.root, pretty_print=True, encoding='utf-8', xml_declaration=True)
        with open(self.__file_path, 'wb') as file:
            file.write(xml_string)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__mode in ['w', 'a']:
            self.__write__()
            self.__save__()


class SSPFile(XMLFile, SSPStandard):

    @abstractmethod
    def __read__(self):
        pass

    @abstractmethod
    def __write__(self):
        pass

    def __check_compliance__(self):
        super().check_compliance(self.schemas[self.identifier], self.namespaces)

    @property
    def identifier(self):
        return self.__identifier

    @property
    def annotations(self):
        return self.__annotations

    def add_annotation(self, annotation: Annotation):
        self.annotations.add_annotation(annotation)

    def __enter__(self):
        register_namespaces()
        return self

    def __init__(self, file_path, mode='r', identifier='unknown'):

        self.__identifier = identifier
        self.__annotations = Annotations()
        super().__init__(file_path, mode)


class SSPElement(ABC):

    @abstractmethod
    def to_element(self) -> ET.Element:
        pass

    @abstractmethod
    def from_element(self, element: ET.Element):
        pass


class EmptyElement(ET.ElementBase):

    def __init__(self, tag, attrib=None, **kwargs):
        super().__init__(tag, attrib, **kwargs)

    def __repr__(self):
        return ET.tostring(self, encoding='utf-8')



import tempfile
import zipfile
import shutil
from pathlib import Path, PosixPath

from pyssp_standard.ssd import SSD
from pyssp_standard.ssb import SSB
from pyssp_standard.ssv import SSV
from pyssp_standard.ssm import SSM
from pyssp_standard.fmu import FMU
from pyssp_standard.standard import SSPStandard


class SSP(SSPStandard):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.__changed:
            temp_resources = Path(self.temp_dir) / self.file_path.stem / 'resources'
            current_resources = list(temp_resources.glob('*'))
            for resource in self.__resources:
                if resource not in current_resources and resource.parent != temp_resources:
                    shutil.copy(resource, temp_resources)
            for resource in current_resources:
                if resource not in self.__resources and resource.parent == temp_resources:
                    resource.unlink()
            filepath = self.file_path.parent / self.file_path.stem
            filepath_zip = Path(str(filepath) + '.zip')
            filepath_ssp = Path(str(filepath) + '.ssp')
            shutil.make_archive(filepath, 'zip', self.temp_dir)
            if filepath_ssp.exists():
                filepath_ssp.unlink()
            filepath_zip.rename(filepath_ssp)
        shutil.rmtree(self.temp_dir)

    def __init__(self, file_path):
        self.__changed = False
        self.temp_dir = tempfile.mkdtemp()
        if type(file_path) is not PosixPath:
            file_path = Path(file_path)
        self.file_path = file_path

        with zipfile.ZipFile(self.file_path, 'r') as zip_ref:
            zip_ref.extractall(self.temp_dir)

        ssp_unpacked_path = Path(self.temp_dir) / self.file_path.stem
        ssp_resource_path = ssp_unpacked_path / 'resources'

        all_resource_files = set(ssp_resource_path.glob('*'))
        self.__ssd = list(ssp_unpacked_path.glob('*.ssd'))[0]
        self.__ssv = list(ssp_resource_path.glob('*.ssv'))
        self.__ssm = list(ssp_resource_path.glob('*.ssm'))
        self.__ssb = list(ssp_resource_path.glob('*.ssb'))
        self.__fmu = list(ssp_resource_path.glob('*.fmu'))

        self.__resources = list(all_resource_files)

    @property
    def ssd(self):
        return SSD(self.__ssd)

    @property
    def ssv(self):
        return [SSV(ssv) for ssv in self.__ssv]

    @property
    def ssm(self):
        return [SSM(file) for file in self.__ssm]

    @property
    def ssb(self):
        return [SSB(file) for file in self.__ssb]

    @property
    def fmu(self):
        return [FMU(file) for file in self.__fmu]

    @property
    def resources(self):
        """ Returns a list of available resources in the ssp folder /resources"""
        return self.__resources

    def add_resource(self, resource):
        """
        Add something to the resource folder of the ssp.
        :param resource: filepath of the object to add.
        """
        self.__changed = True
        self.__resources.append(resource)

    def remove_resource(self, resource_name):
        self.__changed = True
        if type(resource_name) is not str:
            resource_name = resource_name.name
        self.__resources = [path for path in self.__resources if path.name != resource_name]



from pathlib import Path


class SSPStandard:
    namespaces = {'ssc': 'http://ssp-standard.org/SSP1/SystemStructureCommon',
                  'ssv': 'http://ssp-standard.org/SSP1/SystemStructureParameterValues',
                  'ssb': 'http://ssp-standard.org/SSP1/SystemStructureSignalDictionary',
                  'ssm': 'http://ssp-standard.org/SSP1/SystemStructureParameterMapping',
                  'ssd': 'http://ssp-standard.org/SSP1/SystemStructureDescription'}

    __resource_path = Path(__file__).parent / 'resources'
    schemas = {'ssc': __resource_path / 'SystemStructureCommon.xsd',
               'ssd': __resource_path / 'SystemStructureDescription.xsd',
               'ssd11': __resource_path / 'SystemStructureDescription11.xsd',
               'ssm': __resource_path / 'SystemStructureParameterMapping.xsd',
               'ssv': __resource_path / 'SystemStructureParameterValues.xsd',
               'ssb': __resource_path / 'SystemStructureSignalDictionary.xsd'}


class SRMDStandard:
    namespaces = {'stc': 'http://apps.pmsf.net/SSPTraceability/SSPTraceabilityCommon',
                  'ssc': 'http://ssp-standard.org/SSP1/SystemStructureCommon',
                  'srmd': 'http://apps.pmsf.net/STMD/SimulationResourceMetaData'}
    schema = Path(__file__).parent / 'resources' / 'SRMD.xsd'
